/**
 * This Keystore manages two keyslots, a "primary" and a "secondary" slot
 * The key locations (0 and 1) are at EEPROM location EEP_KEY_OFFSET and EEP_KEY_OFFSET+sizeof(KEY)
 * The value at EEP_PRIMARY_KEY_SLOT declares which keyslot (0 or 1) ist the "primary" slot.
 * The remaining one is the "secondary"
 * 
 * If the "primary" key is requested, the value at EEP_PRIMARY_KEY_SLOT is looked up and the key is read from the indicated location
 * If the "secondary" key is requested the key is read from the alternate location.
 * 
 * When the "primary" key is updated (parameter "update_secondary" = false) the "secondary" keyslot is overwritten
 * with the new key and then declared as "primary" by flipping the value stored at EEP_PRIMARY_KEY_SLOT.
 * This procedure ensures that there is always one valid key in the store.
 * 
 * When the "secondary" key is updated (parameter "update_secondary" = false) the "primary" keyslot is overwritten
 * with the new key. The value at EEP_PRIMARY_KEY_SLOT stays unchanged, declaring the new key as "primary".
 * The old secondary key stays at place.
 * This procedure ensures that there is always one valid key in the store.
 */

#include "keystore.h"


#define EEP_PRIMARY_KEY_SLOT 127
#define EEP_KEY_OFFSET 128

static uint8_t get_keyslot_index(uint8_t secondary);
static void write_new_primary_key(uint8_t* key, uint8_t slotindex);
void xor(uint8_t* src1dst, uint8_t* src2, uint8_t length);

uint8_t keybuffer[sizeof(KEY)];

uint8_t keystore_is_empty(void){
	return (eeprom_read_byte((const uint8_t*)EEP_PRIMARY_KEY_SLOT) == 0xFF);
}
uint8_t* keystore_get_key(int8_t secondary){
	uint8_t slotindex = get_keyslot_index(secondary);
	
	eeprom_read_block(keybuffer, (const void*)EEP_KEY_OFFSET + (slotindex*sizeof(KEY)), sizeof(KEY));
	return keybuffer;
}

void keystore_init_key(uint8_t* key){
	uint8_t slotindex = get_keyslot_index(0);
	write_new_primary_key(key, slotindex);
}

void keystore_update_key(uint8_t* new_key_delta, uint8_t update_secondary){
	uint8_t read_slotindex = get_keyslot_index(update_secondary); //read from the valid slot
	
	eeprom_read_block(keybuffer, (const void*)EEP_KEY_OFFSET + (read_slotindex*sizeof(KEY)), sizeof(KEY));
	xor(keybuffer, new_key_delta, sizeof(KEY));
	
	uint8_t write_slotindex = get_keyslot_index(!update_secondary); //write to the invalid slot
	write_new_primary_key(keybuffer, write_slotindex);
}

void keystore_clear(void){
	for (uint8_t i = 0; i < sizeof(KEY)*2; i++){
		eeprom_write_byte((uint8_t*)(EEP_KEY_OFFSET+i), 0xFF);
	}
	eeprom_write_byte((uint8_t*)EEP_PRIMARY_KEY_SLOT, 0xFF);
}

static uint8_t get_keyslot_index(uint8_t secondary){
	uint8_t primary_slot = eeprom_read_byte((const uint8_t*)EEP_PRIMARY_KEY_SLOT);
	if (primary_slot > 1) primary_slot = 0;
	return secondary ? !primary_slot : primary_slot;
}
static void write_new_primary_key(uint8_t* key, uint8_t slotindex){
	eeprom_write_block(key, (void*)EEP_KEY_OFFSET + (slotindex*sizeof(KEY)) , sizeof(KEY));
	eeprom_write_byte((uint8_t*)EEP_PRIMARY_KEY_SLOT, slotindex);
}
void xor(uint8_t* src1dst, uint8_t* src2, uint8_t length){
	for (uint8_t i = 0; i < length; i++){
		src1dst[i] ^= src2[i];
	}
}
