#include <string.h>
#include "portal.h"
#include "../../ISO7816-card.h"
#include "keystore.h"
#include "../../rng.h"

#define PORTAL_INS_QUERY_VERSION 0x00
#define PORTAL_INS_INIT_KEY      0x01
#define PORTAL_INS_AUTH          0x02
#define PORTAL_INS_UPDATE_KEY    0x03

uint16_t portal_handle_adpu(ISO7816_APDU_Header* header,
							 uint8_t* payload,
							 uint8_t payload_length,
							 uint8_t* answer_buffer,
							 uint8_t* answer_length);

sha1_ctx_t sha1_state;
uint8_t challenge[sizeof(SALT)];
uint8_t current_auth_key = 1;

void portal_init(void){
	rnd_get_bytes(challenge, sizeof(SALT));
}

apdu_handler_t portal_select(void){
	return portal_handle_adpu;
}

uint16_t portal_handle_adpu(ISO7816_APDU_Header* header,
							 uint8_t* payload,
							 uint8_t payload_length,
							 uint8_t* answer_buffer,
							 uint8_t* answer_length){
	
	switch (header->instruction){
		case PORTAL_INS_QUERY_VERSION:
			if (*answer_length < 1) return ISO7816_STATUS_ERR_WRONG_LENGTH;
			
			*answer_length = 1;
			answer_buffer[0] = 2;
			return ISO7816_STATUS_OK;
			break;
			
			
		case PORTAL_INS_INIT_KEY:
			if (payload_length != sizeof(KEY)) return ISO7816_STATUS_ERR_WRONG_LENGTH + payload_length+1;
			
			*answer_length = 0;
			if (keystore_is_empty()){
				keystore_init_key(payload);
				return ISO7816_STATUS_OK;
			} else {
				return ISO7816_STATUS_ERR_CONDITIONS_NOT_SATISFIED;
			}
			break;
			
			
		case PORTAL_INS_AUTH:
			if (*answer_length < sizeof(HASH) + sizeof(SALT)) return ISO7816_STATUS_ERR_WRONG_LENGTH;
			if (payload_length != sizeof(SALT)) return ISO7816_STATUS_ERR_WRONG_LENGTH;
			
			current_auth_key = !current_auth_key;
			*answer_length = sizeof(HASH) + sizeof(SALT);
			sha1_init(&sha1_state);
			sha1_nextBlock(&sha1_state, keystore_get_key(current_auth_key));
			sha1_lastBlock(&sha1_state, payload, sizeof(SALT)*8);
			sha1_ctx2hash(answer_buffer, &sha1_state);
			memcpy(answer_buffer + sizeof(HASH), challenge, sizeof(SALT));
			
			return ISO7816_STATUS_OK;
			break;
			
			
		case PORTAL_INS_UPDATE_KEY:
			if (payload_length != sizeof(KEY)+ sizeof(HASH)) return ISO7816_STATUS_ERR_WRONG_LENGTH;
			*answer_length = 0;
			
			uint8_t* new_key_delta = payload;
			uint8_t* challange_response = payload + sizeof(KEY);
			
			sha1_init(&sha1_state);
			sha1_nextBlock(&sha1_state, keystore_get_key(current_auth_key));
			sha1_lastBlock(&sha1_state, challenge, sizeof(SALT)*8);
			if (! sha1_ctx_hash_compare(challange_response, &sha1_state)){
				return ISO7816_STATUS_ERR_SECURITY_STATUS_NOT_SATISFIED;
// 				*answer_length = 20;
// 				memcpy(answer_buffer, (const void*)challange_response, 20);
// 				return ISO7816_STATUS_OK;
			}
			
			
			keystore_update_key(new_key_delta, current_auth_key);
			
			return ISO7816_STATUS_OK;
			break;
			
			
		case 0xFE:
			if (*answer_length < sizeof(KEY)) return ISO7816_STATUS_ERR_WRONG_LENGTH;
			
			memset(answer_buffer, 0xBB, sizeof(KEY));
			memcpy(answer_buffer, keystore_get_key(0), sizeof(KEY));
			if (*answer_length < 2*sizeof(KEY)){
				*answer_length = sizeof(KEY);
			} else {
				memcpy(answer_buffer + sizeof(KEY), keystore_get_key(1), sizeof(KEY));
				*answer_length = 2*sizeof(KEY);
			}
			return ISO7816_STATUS_OK;
			break;
			
			
		case 0xFF:
			*answer_length = 0;
			keystore_clear();
			
			return ISO7816_STATUS_OK;
			break;
	}
	return ISO7816_STATUS_ERR_INS_NOT_SUPPORTED;
}

void portal_deselect(void){
	
}
