#include "software_serial.h"

#ifdef UART_USE_AS_STDOUT //Sets up the serial port as a stream for stdout. With this printf and stuff will write to the serial port.
	#include <stdio.h>
	
	void UART_transmit_stdout(char c, FILE* stream){
		UART_transmit_char((unsigned char)c);
	}

    static FILE UART_stdout = FDEV_SETUP_STREAM(UART_transmit_stdout, NULL, _FDEV_SETUP_WRITE);
#endif


/**
 * Initializes the Software UART
 */
void UART_init(void){
	#ifdef UART_USE_AS_STDOUT
		stdout = &UART_stdout; //assign serial output stream as stdout.
	#endif
	
	UART_recv_init();
}

/**
 * Switches the UART to transmit mode
 */
void UART_transmit_init(void){
	cli(); //interrupts disable (messes up send timings otherwise)
	SWSERIAL_PORT |= (1 << SWSERIAL_LINE); //data line to high level (otherwise we might trigger a start bit)
	SWSERIAL_PORT_DR |= (1 << SWSERIAL_LINE); //output
}

/**
 * Sends a 0-terminated string
 * Note: The caller must to switch to transmit mode before calling this
 */
void UART_transmit_buffer_sz(char* buffer){
	while (buffer[0] != 0){
		UART_transmit_char(buffer[0]);
		buffer++;
	}
}

/**
 * Sends a byte coverted to hex digits
 * Note: The caller must to switch to transmit mode before calling this
 */
void UART_transmit_byte_hex(unsigned char b){

	unsigned char n;
	n = '0' + (b >> 4 & 0xF);
	UART_transmit_char((n > '9') ? n + 7 : n);
	n = '0' + (b & 0xF);
	UART_transmit_char((n > '9') ? n + 7 : n);
}


/**
 * Writes a single character to the data line, bit-by-bit
 * Note: The caller must to switch to transmit mode before calling this
 */
void UART_transmit_char(unsigned char x){

	_delay_us(BIT_LENGTH); //guard time beween bytes
	
	WRITE(0); //start bit
	
	#if SWSERIAL_PARITY == EVEN
		uint8_t parity = 0;
	#endif
	#if SWSERIAL_PARITY == ODD
		uint8_t parity = 1;
	#endif
	for (unsigned char i = 0; i < 8; i++){
		_delay_us(BIT_LENGTH - INST_LENGTH(8));
		WRITE(x & 1);
		#if SWSERIAL_PARITY != NONE
			parity ^= x & 1;
		#endif
		x >>= 1;
	}
	
	#if (SWSERIAL_PARITY != NONE)
		_delay_us(BIT_LENGTH - INST_LENGTH(8));
		WRITE(parity);		
	#endif
	
	_delay_us(BIT_LENGTH - INST_LENGTH(8));
	WRITE(1); //stop bit
	
	#if SWSERIAL_STOPBITS == 2
	_delay_us(BIT_LENGTH - INST_LENGTH(8));
	WRITE(1); //stop bit 2
	#endif
	
	_delay_us(BIT_LENGTH); //guard time beween bytes
	_delay_us(BIT_LENGTH); //guard time beween bytes
}


//------------------------- Receive code ----------------------------

uint8_t recv_buffer1 = 0;          //Byte currently beeing received (only used by timer interrupt handler)
uint8_t recv_bits = 0;             //Number of bits done in current receive (only used by timer interrupt handler)
volatile uint8_t recv_buffer2 = 0; //Last Byte that finished receiving
volatile uint8_t recv_avail = 0;   //Is a byte ready to be processed?
volatile uint8_t recv_parity;
volatile uint8_t recv_valid = 0;    //Did we receive 1 for the stop bits and valid parity

/**
 * Switch UART into receive mode
 */
void UART_recv_init(void){
	SWSERIAL_PORT_DR &= ~(1 << SWSERIAL_LINE); //set Serial pin to input
	SWSERIAL_PORT |= (1 << SWSERIAL_LINE); //pullup
	
	DDRB |= (1 << 5) | (1 << 7);
	
	TIFR = (1 << ICF1); //clear old interrupt flag.
	TIMSK = 1 << TICIE1; //enable ICP interrupt, disable timer compare interrupt
	TCCR1B =  (1 << CS10)   //Timer prescaler = 1
// #ifdef CTC1
			| (1 << CTC1)   //Clear timer on compare match (important for exact timing, don't do it in software)
// #else
// 			| (1 << WGM12)
// #endif
			| (1 << ICNC1) //Capture interrupt noise canceler enable (filters out very short spikes  < 4 clks)
// #ifdef ICES1
// 			| (1 << ICES1)
// #endif
			;
	
	OCR1A  = 372; //Default smartcard conversion factor 3571200Hz/372(clks/bit) = 9600bit/s
	
	sei(); //global interrupt enable
}

/**
 * Is a byte available to be processed?
 */
inline uint8_t UART_data_available(void){
	return recv_avail != 0;
}
/**
 * Reads the last byte that was received.
 * Resets the availability flag.
 */
uint8_t UART_read_char(void){
	recv_avail = 0;
	return recv_buffer2;
}

uint16_t UART_receive_char_timeout(uint16_t timeout){
	while ((timeout > 0) && (recv_avail == 0)) {_delay_us(100); timeout--;}
	if (recv_avail == 0) return UART_ERR_TIMEOUT;
	
	return UART_read_char();
}

/**
 * Enables the timer (interrupt) that samples the data line to receive the bits.
 * Disables the Capture (ICP) interrupt that triggered on the start bit.
 */
void UART_start_read(void){
	recv_bits = 0;
	
	#if SWSERIAL_PARITY == EVEN
		recv_parity = 0;
	#endif
	#if SWSERIAL_PARITY == ODD
		recv_parity = 1;
	#endif
	
	TCNT1 = 0xFFFF-70; //Timer value. We start above the match counter, instead of starting af 0, this makes the first tick take a little longer, moving the sampling further inside the bit, not at the very beginning.

	TIFR = (1 << OCF1A); //clear old interrupt flag
	TIMSK = (1 << OCIE1A); //Disable ICP interrupt, Enable timer Compare Interrupt.
}

/**
 * Capture interrupt, trigered by a falling edge on the data line (start bit)
 */
ISR(TIMER1_CAPT_vect){
	PORTB |= (1 << 5); //Debug
	UART_start_read();
	PORTB &= ~(1 << 5); //Debug
}

/**
 * Timer interrupt, samples the data line and writes the bits to the recv buffer.
 * If all bits are read it returns control to the Capture interrupt and disables itself.
 */
ISR(TIMER1_COMPA_vect){
	PORTB |= (1 << 7); //Debug
	
	uint8_t bit = (SWSERIAL_PIN & (1 << SWSERIAL_LINE)) ? 1 : 0;
	
	if (recv_bits < 8){
		recv_buffer1 >>= 1;
		recv_buffer1 |= (bit ? 0x80 : 0);
		if (bit) recv_parity = !recv_parity;
	}
	if (recv_bits == 8){
		recv_valid = bit == recv_parity;
	}
	if (recv_bits == 9){
		if (!bit) recv_valid = 0;
	}
	
	recv_bits++;
	if (recv_bits == 10){
		if (recv_valid){
			recv_buffer2 = recv_buffer1;
			recv_avail = 1;
		}
		recv_bits = 0;
		UART_recv_init();
	}
	PORTB &= ~(1 << 7); //Debug
}
