#include "common.h"
#include "ISO7816-reader.h"

#include "clockout.h"

#include <util/delay.h>
#include <avr/wdt.h>
#include <avr/sleep.h>

#include "lcd.h"

#include <avr/pgmspace.h>
#include <string.h>

#include "../common/transport/transport.h"
#include "../common/cardreader_interface-common.h"


volatile uint8_t card_status = CARD_STATUS_REMOVED;

volatile uint8_t powersave_mode = 1;
uint8_t card_atr_error = 0;
ISO7816_ATR card_atr;

#define HOST_TIMEOUT 200
#define HOST_PING_TIME 100
volatile uint16_t host_timeout = HOST_TIMEOUT;
volatile uint8_t host_ping = 1;
volatile uint8_t host_dead = 0;

#define DISPLAY_RESET_TIMEOUT 1000 /* 1000*10 = 4 sec */
volatile uint16_t display_reset_timeout = DISPLAY_RESET_TIMEOUT;
volatile uint8_t  display_reset = 1;

//----------- Hardware interrupt handler -----------
void ISO7816_card_presence_changed(uint8_t inserted){

	if (inserted){
		ISO7816_card_power(1);
		clockout_start();
		card_status = CARD_STATUS_INSERTED;
	} else {
		clockout_stop();
		ISO7816_card_power(0);
		card_status = CARD_STATUS_REMOVED;
	}
}

ISR(TIMER0_COMPA_vect){
	if (display_reset_timeout){
		display_reset_timeout--;
		if (display_reset_timeout == 0){
			display_reset = 1;
		}
	}
	if (host_timeout){
		host_timeout--;
		if (host_timeout == HOST_PING_TIME || host_timeout == 0){
			if (host_timeout == 0){
				host_dead = 1;
				display_reset = 1;
				host_timeout = HOST_TIMEOUT;
			}
			host_ping = 1;
		}
	}
}

void timer_init(void){
	TCCR0A = (1 << WGM01);
	TCCR0B = (1 << CS00) | (1 << CS02); //Prescaler /1024
	OCR0A = 89; // 9216000Hz/1024/(1+89) = 100Hz
	TIMSK0 = (1 << OCIE0A);
}

//----------- Interrupt Message Handler -----------
void reset(void){
	wdt_enable(WDTO_15MS);
	while(1){;}
}
void host_alive(void){
	host_timeout = HOST_TIMEOUT;	
	if (host_dead){
		host_dead = 0;
		display_reset = 1;
	}
}
void cardreader_alive(void){
	transport_send_interrupt_message(CARDREADER_INTERRUPT_CARDREADER_ALIVE);
}

//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

void handle_sys_msg(uint8_t msg){
	uint8_t ret;
	switch (msg){
		case CARDREADER_SYS_MSG_GET_CARD_STATUS:
			transport_send_message((uint8_t*)&card_status, 1);
			break;
		case CARDREADER_SYS_MSG_GET_CARD_ATR:
			if (card_status != CARD_STATUS_READY){
				transport_send_message((uint8_t*)&card_atr, sizeof(ISO7816_ATR));
			} else {
				transport_send_message(&card_atr_error, 1);
			}
				
		case CARDREADER_SYS_MSG_IS_DISPLAY_CONNECTED:
			//TODO: Come up with a clever way to figure that out.
			ret = 1;
			transport_send_message(&ret, 1);
			break;
	}
}

void handle_card_msg(uint8_t* msg, uint8_t length){
	uint8_t answer_buffer[0x7F+2];
	ISO7816_APDU_Header* apdu_header         = (ISO7816_APDU_Header*)msg;
	uint8_t*             apdu_payload        = msg    + sizeof(ISO7816_APDU_Header);
	uint8_t              apdu_payload_length = length - sizeof(ISO7816_APDU_Header);
	uint16_t*            answer_status = (uint16_t*)answer_buffer; //status code first
	uint8_t*             answer_payload = answer_buffer + 2;       //reply afterwards
	uint8_t              answer_payload_length = 0x7F; //this is the max we can send over transport (127+2 status bytes).
	                                                   //we need to allocate this buffer size anyway,
	                                                   //so why care adding a LE field in the card message?
	
	uint16_t result = ISO7816_send_apdu(apdu_header,
						   apdu_payload,
						   apdu_payload_length,
						   answer_status,
						   answer_payload,
						   answer_payload_length);
	if (ISO7816_IS_ERROR(result)){
		//send reader error code (they are distinct from ISO7816 status words, so there is no collision)
		transport_send_message((uint8_t*)&result, 2);
	} else {
		answer_payload_length = result & 0xFF;
		
		transport_send_message(answer_buffer, answer_payload_length+2); //send the card status(2byte) and answer_payload
	}
}

void handle_display_msg(uint8_t* msg, uint8_t length){
	uint8_t type = msg[0];
	uint8_t success = 0;
	
	char* text = (char*)(&msg[1]);
	uint8_t text_len = length - 1;
	
	display_reset_timeout = DISPLAY_RESET_TIMEOUT;
	
	switch (type){
		case CARDREADER_DISPLAY_MSG_CLEAR:
			lcd_clear();
			lcd_home();
			success = 1;
			break;
		case CARDREADER_DISPLAY_MSG_SET_POS:
			if (length == 3){
				lcd_set_pos(msg[1], msg[2]);
				success = 1;
			}
			break;
		case CARDREADER_DISPLAY_MSG_WRITE:
			
			if (text_len <= 16){
				lcd_write_string(text, text_len);
			} else {
				lcd_clear();
				lcd_home();
				lcd_write_string(text, 16);
				lcd_set_pos(0,1);
				lcd_write_string(text+16, text_len - 16);
			}
			success = 1;
			break;
		case CARDREADER_DISPLAY_MSG_SET_BACKLIGHT:
			if (length == 2){
				lcd_set_backlight(msg[1]);
				success = 1;
			}
			break;
	}
	transport_send_message(&success, 1);
}

void display_receive_error(uint16_t status){
// 	lcd_write_char('e');
// 	lcd_write_byte_hex(status);
	lcd_clear();
	lcd_home();
	lcd_write_string_sz_P(PSTR("Host Read Error"));
	lcd_set_pos(0,1);
	switch (status){
		case 0:                              lcd_write_string_sz_P(PSTR("0-Byte Message")); break;
		case TRANSPORT_ERR_TIMEOUT:          lcd_write_string_sz_P(PSTR("Read timeout")); break;
		case TRANSPORT_ERR_CHECKSUM_FAIL:    lcd_write_string_sz_P(PSTR("Checksum fail")); break;
// 		case TRANSPORT_ERR_MESSAGE_ABORT:    lcd_write_string_sz_P(PSTR("Msg truncated")); break;
		case TRANSPORT_ERR_MESSAGE_TOO_LONG: lcd_write_string_sz_P(PSTR("Msg too long")); break;
// 		case TRANSPORT_ERR_UNEXPCETED_END:   lcd_write_string_sz_P(PSTR("Unexp End of Msg")); break;
// 		case TRANSPORT_ERR_NESTED_INTERRUPTS:lcd_write_string_sz_P(PSTR("Nested int. Msgs")); break;
	}
}

void display_card_status(void){
	lcd_clear();
	lcd_home();
	if (host_dead){
		lcd_write_string_sz_P(PSTR("0xDEAD 0xBEAF"));
		return;
	}
	
	lcd_write_string_sz_P(PSTR("Warpzone Portal"));
	lcd_set_pos(0,1);
	switch (card_status){
		case CARD_STATUS_REMOVED:
			lcd_write_string_sz_P(PSTR("Insert Card"));
			break;
		case CARD_STATUS_INSERTED:
			lcd_write_string_sz_P(PSTR("Reading Card"));
			break;
		case CARD_STATUS_READY:
			lcd_write_string_sz_P(PSTR("Card Ready"));
			break;
		case CARD_STATUS_ERROR:
// 			lcd_write_string_sz_P(PSTR("Card Error 0x"));
// 			lcd_write_byte_hex(card_atr_error);
// 			lcd_write_char(' ');
			switch(card_atr_error){
				case ERR_ATR_TIMEOUT:                  lcd_write_string_sz_P(PSTR("Card Mute")); break;
				case ERR_TOO_MANY_ATR_INTERFACE_BYTES: lcd_write_string_sz_P(PSTR("ATR too long")); break;
				case ERR_ATR_CHECKSUM_FAIL:            lcd_write_string_sz_P(PSTR("ATR chksum fail")); break;
				case ERR_PROTOCOL_NOT_SUPPORTED:       lcd_write_string_sz_P(PSTR("Card unspported")); break;
				case ERR_ATR_CARD_REMOVED:             lcd_write_string_sz_P(PSTR("Card removed")); break;
				default:                               lcd_write_string_sz_P(PSTR("Unknown error ")); lcd_write_byte_hex(card_atr_error); break;
			}
			break;
	}
// 	lcd_write_char('C');
// 	lcd_write_byte_hex(card_status);
}

void sleep(void){
	transport_flush();//be sure all data is sent to the host.
	wdt_disable(); //Disable watchdog while sleeping
	
// 	set_sleep_mode(SLEEP_MODE_IDLE);
// 	sleep_mode();
// 	sleep_disable();
	
	wdt_enable(WDTO_8S);
}

int main(void){	
	wdt_reset();
	wdt_enable(WDTO_8S);
	
	ACSR |= (1 << ACD); //disable Analog comparator. don't need it.
	PRR |= (1 << PRADC) | (1 << PRSPI); //disable ADC and SPI as well.
	
	ISO7816_init();
	

	transport_init();
	transport_set_interrupt_handler(CARDREADER_INTERRUPT_RESET, reset);
	transport_set_interrupt_handler(CARDREADER_INTERRUPT_HOST_ALIVE, host_alive);
	transport_set_interrupt_handler(CARDREADER_INTERRUPT_CARDREADER_ALIVE, cardreader_alive);
	
	lcd_set_backlight(0);
	_delay_ms(200);
	lcd_init();
	
	display_card_status();
	//lcd_write_string_sz_P(PSTR("Waiting for Host"));

	
	if (ISO7816_is_card_present()) card_status = CARD_STATUS_INSERTED;
	
	timer_init();
	sei();
	
	cardreader_message_t msg;
	uint16_t recv_status;
	uint8_t length;
	
	uint8_t old_card_status = CARD_STATUS_REMOVED;

	transport_send_interrupt_message(CARDREADER_INTERRUPT_CARDREADER_ALIVE);
	
	while(1){
		wdt_reset();
		
		if (display_reset){
			display_reset = 0;
			display_card_status();
		}
		if (host_ping){
			transport_send_interrupt_message(CARDREADER_INTERRUPT_HOST_ALIVE);
			host_ping = 0;
		}
		
		if (card_status == CARD_STATUS_REMOVED) sleep();
		
		if (card_status != old_card_status){
			old_card_status = card_status;
			switch(card_status){
				case CARD_STATUS_REMOVED:
					clockout_stop();
					ISO7816_card_power(0);
					memset(&card_atr, 0,  sizeof(ISO7816_ATR));
					card_atr_error = 0;
					
 					display_card_status();
					transport_send_interrupt_message(CARDREADER_INTERRUPT_CARD_REMOVED);
					break;
				case CARD_STATUS_INSERTED:
 					lcd_home();
 					lcd_write_string_sz_P(PSTR("Reading card... "));
					ISO7816_card_power(1);
					clockout_start();
					uint8_t ret = ISO7816_readATR(&card_atr);
					if (ATR_IS_ERROR(ret)){
						//Set to error state (if the error was not caused by a removed card)
						if (ret != ERR_ATR_CARD_REMOVED) card_status = CARD_STATUS_ERROR;
					} else {
						card_status = CARD_STATUS_READY;
					}
					
					card_atr_error = ret;
					old_card_status = card_status;
					
 					display_card_status();
					if (card_status == CARD_STATUS_READY) transport_send_interrupt_message(CARDREADER_INTERRUPT_CARD_INSERTED);
					break;
				default:
					display_card_status();
			}
		}
		
		if (transport_data_available()){
			recv_status = transport_receive_message((uint8_t*)&msg, CARDREADER_MAX_MSG_LEN, 30000);
			if (TRANSPORT_IS_ERROR(recv_status) || recv_status == 0){//Handle error codes or zero length msg (invalid here)
				display_receive_error(recv_status);
			} else {
				length = (recv_status & 0xFF) - 1; //if its not an error code is the msg length < 256. Subtract 1 for the type byte
// 				lcd_home();
// 				lcd_write_string_sz_P(PSTR("Msg:"));
// 				lcd_write_byte_hex(msg.type);
// 				lcd_write_string_sz_P(PSTR(" "));
// 				lcd_write_byte_hex(length);
				switch (msg.type){
					case CARDREADER_MSG_TYPE_SYS:
						handle_sys_msg(msg.body[0]);
						break;
					case CARDREADER_MSG_TYPE_CARD:
						handle_card_msg(msg.body, length);
						break;
					case CARDREADER_MSG_TYPE_DISPALY:
						handle_display_msg(msg.body, length);
						break;
				}
			}
		}
	}
	
}