#include "keystore.h"

//Forward declares
void keystore_update_hashes(SALT salt);
uint8_t keystore_read_key(uint16_t index, KEY key);
uint8_t keystore_write_key(uint16_t index, KEY key);
uint8_t keystore_read_backup_key(uint16_t index, KEY key);
uint8_t keystore_write_backup_key(uint16_t index, KEY key);
uint8_t keystore_eeprom_read(uint8_t eeprom, uint16_t offset, uint8_t* buffer, uint8_t length);
uint8_t keystore_eeprom_write(uint8_t eeprom, uint16_t offset, uint8_t* buffer, uint8_t length);

void xor(uint8_t* src1dest, uint8_t* src2, uint8_t length);

static SALT salt;
static uint8_t salt_usage_count = 0;
HASH_HEAD hashes[KEY_COUNT];

void keystore_init(){
	i2c_master_init();
	rnd_init();
	keystore_update_salt();
	
	//Enable EEPROM write protect
	EEPROM_WRITE_PROTECT_DDR  |= (EEPROM_WRITE_PROTECT_MAIN | EEPROM_WRITE_PROTECT_BACKUP);
	EEPROM_WRITE_PROTECT_PORT |= (EEPROM_WRITE_PROTECT_MAIN | EEPROM_WRITE_PROTECT_BACKUP);
}

uint8_t* keystore_get_salt(){
	salt_usage_count++;
	//We allow the same salt to be queried up to 3 times
	//for retries in case of transfer errors or the card switching to the backup key.
	//Ater 3 retries a new salt is generated. This takes quite some time.
	if (salt_usage_count > 3) keystore_update_salt();
	return salt;
}

int16_t keystore_find_key(HASH keyhash){
	HASH refhash;
	for (uint16_t index = 0; index < KEY_COUNT; index++){
		
		//TODO: No need to compare for empty keyslots.
		if (memcmp(&(hashes[index]), keyhash, sizeof(HASH_HEAD)) == 0){
			if (keystore_read_slot_status(index) == KEYSLOT_USED){
				if (keystore_calc_keyhash(index, salt, refhash)){
					if (memcmp(refhash, keyhash, sizeof(HASH)) == 0){
						return index;
					}
				}
			}
		}
	}
	return -1;
}

void keystore_update_salt(){
	rnd_get_bytes(salt, sizeof(SALT));
	keystore_update_hashes(salt);
	salt_usage_count = 0;
}

void keystore_gen_keydelta(KEY keydelta){
	rnd_get_bytes(keydelta, sizeof(KEY));
}

uint8_t keystore_update_slot(uint16_t index, KEY keydelta){
	KEY key;
	if (!keystore_read_key(index, key)) return 0;
	keystore_write_backup_key(index, key); //Backup old key.
	xor(key, keydelta, sizeof(KEY));
	if (!keystore_write_key(index, key)) return 0;
	return 1;
}

void keystore_update_hashes(SALT salt){
	uint8_t buffer[sizeof(KEY) + sizeof(SALT)];
	HASH fullhash;
	
	memcpy(buffer + sizeof(KEY), salt, sizeof(SALT));
	
	for (uint16_t index = 0; index < KEY_COUNT; index++){
		wdt_reset(); //this may take a while, keep the Watchdog happy.
		
#ifdef KEYSTORE_DEBUG
		printf_P(PSTR("Debug: Slot %03i: "), index);
#endif
		uint8_t slot_status = keystore_read_slot_status(index);
		if (slot_status == KEYSLOT_USED || slot_status == KEYSLOT_DISABLED){
			keystore_read_key(index, buffer);
			
#ifdef KEYSTORE_DEBUG
			printf_P(PSTR("K:"));
			MGMT_transmit_buffer_hex(buffer, 8);
#endif
			
			sha1(fullhash, buffer, (sizeof(KEY) + sizeof(SALT))*8);
			memcpy(&(hashes[index]), fullhash, sizeof(HASH_HEAD));
			
#ifdef KEYSTORE_DEBUG
			printf_P(PSTR(" H:"));
			MGMT_transmit_buffer_hex(&hashes[index], sizeof(HASH_HEAD));
			
			if (slot_status == KEYSLOT_DISABLED){
				printf_P(PSTR(" (Disabled)"));
			}
			
			printf_P(PSTR("\n"));
			
		} else if (slot_status == KEYSLOT_EMPTY){
			printf_P(PSTR("Empty\n"));
		} else if (slot_status == KEYSLOT_ERROR){
			printf_P(PSTR("Read Error\n"));
		} else {
			printf_P(PSTR("Unknown (%02x)\n"), slot_status);
#endif
		}
	}
}

void keystore_list(void){
	for (uint16_t index = 0; index < KEY_COUNT; index++){
		wdt_reset(); //this may take a while, keep the Watchdog happy.
		
		printf_P(PSTR("Slot %03i: "), index);
		uint8_t slot_status = keystore_read_slot_status(index);
		if (slot_status == KEYSLOT_USED || slot_status == KEYSLOT_DISABLED){
#ifdef KEYSTORE_DEBUG
			KEY key;
			
			keystore_read_key(index, key);
			
			printf_P(PSTR("K:"));
			MGMT_transmit_buffer_hex(key, 8);
			
			printf_P(PSTR(" H:"));
			MGMT_transmit_buffer_hex(&(hashes[index]), sizeof(HASH_HEAD));
#else
			printf_P(PSTR("Used"));
#endif
			if (slot_status == KEYSLOT_DISABLED){
				printf_P(PSTR(" (Disabled)"));
			}
			
			printf_P(PSTR("\n"));
			
		} else if (slot_status == KEYSLOT_EMPTY){
			printf_P(PSTR("Empty\n"));
		} else if (slot_status == KEYSLOT_ERROR){
			printf_P(PSTR("Read Error\n"));
		} else {
			printf_P(PSTR("Unknown (%02x)\n"), slot_status);
		}
	}
}

uint8_t keystore_calc_keyhash(uint16_t index, SALT salt, HASH keyhash){
	uint8_t buffer[sizeof(SALT)+sizeof(KEY)];
	uint8_t ret = keystore_read_key(index, buffer);
	if (ret == 0) return 0;
	memcpy(buffer+sizeof(KEY), salt, sizeof(SALT));
#ifdef KEYSTORE_DEBUG
	printf_P(PSTR("sha1("));
	MGMT_transmit_buffer_hex(buffer, sizeof(SALT)+sizeof(KEY));
	printf_P(PSTR(")\n"));
#endif
	sha1(keyhash, (uint8_t*)buffer, (sizeof(SALT)+sizeof(KEY))*8);
	return 1;
}

uint8_t keystore_restore_slot(uint16_t index){
	KEY key;
	if (!keystore_read_backup_key(index, key)) return 0;
	if (!keystore_write_key(index, key))       return 0;
	
	uint8_t status;
	if (!keystore_eeprom_read(KEY_BACKUP_EEPROM, EEPROM_TABLE_OFFSET + index, &status, sizeof(uint8_t))) return 0;
	if (!keystore_eeprom_write(KEY_MAIN_EEPROM,  EEPROM_TABLE_OFFSET + index, &status, sizeof(uint8_t))) return 0;
	
	if (status == KEYSLOT_USED || status == KEYSLOT_DISABLED){
		keystore_calc_keyhash(index, salt, (uint8_t*)&hashes[index]);
	}
	
	return 1;
}

void xor(uint8_t* src1dest, uint8_t* src2, uint8_t length){
	for (uint8_t i = 0; i < length; i++){
		src1dest[i] ^= src2[i];
	}
}

uint16_t keystore_find_empty_slot(void){
	for (uint16_t index = 0; index < KEY_COUNT; index++){
		wdt_reset(); //this may take a while, keep the Watchdog happy.
		
		uint8_t slot_status = keystore_read_slot_status(index);
		if (slot_status == KEYSLOT_EMPTY) return index;
	}
	return 0xFFFF;
}

#include <util/delay.h>

/**
 * Initializes the given keyslot with a new random key
 * @param index index to the keyslot to initialize
 * @param keyout buffer where the new key will be written to. MUST be at least 64 (sizeof(KEY)) bytes long
 * @return 1 if successful, 0 if not.
 */
uint8_t keystore_init_slot(uint16_t index, KEY keyout){
	if (!rnd_get_bytes(keyout, sizeof(KEY))) return 0;
	if (!keystore_write_slot_status(index, KEYSLOT_EMPTY)) return 0;
	_delay_us(40);
	if (!keystore_write_key(index, keyout)) return 0;
	_delay_us(40);
	if (!keystore_write_slot_status(index, KEYSLOT_USED)) return 0;
	_delay_us(40);
	
	sha1_ctx_t s;
	sha1_init(&s);
	sha1_nextBlock(&s, keyout);
	sha1_lastBlock(&s, salt, sizeof(SALT));
	sha1_ctx2hashhead(&hashes[index], &s);
	
	return 1;
}


uint8_t keystore_clear_slot(uint16_t index){
	if (!keystore_write_slot_status(index, KEYSLOT_EMPTY)) return 0;
	_delay_us(40);
	return 1;
}
uint8_t keystore_disable_slot(uint16_t index){
	if (keystore_read_slot_status(index) != KEYSLOT_USED) return 0;
	
	if (!keystore_write_slot_status(index, KEYSLOT_DISABLED)) return 0;
	_delay_us(40);
	return 1;
}
uint8_t keystore_enable_slot(uint16_t index){
	if (keystore_read_slot_status(index) != KEYSLOT_DISABLED) return 0;
	
	if (!keystore_write_slot_status(index, KEYSLOT_USED)) return 0;
	_delay_us(40);
	return 1;
}

uint8_t keystore_read_slot_status(uint16_t index){
	uint8_t key_status = KEYSLOT_ERROR;
	uint8_t ret = keystore_eeprom_read(KEY_MAIN_EEPROM, EEPROM_TABLE_OFFSET + index, &key_status, sizeof(uint8_t));
	return (ret == 1) ? key_status : KEYSLOT_ERROR;
}

uint8_t keystore_write_slot_status(uint16_t index, uint8_t key_status){
	uint8_t ret = keystore_eeprom_write(KEY_MAIN_EEPROM,  EEPROM_TABLE_OFFSET + index, &key_status, sizeof(uint8_t));
	keystore_eeprom_write(KEY_BACKUP_EEPROM,  EEPROM_TABLE_OFFSET + index, &key_status, sizeof(uint8_t));
	return (ret == 1) ? key_status : KEYSLOT_ERROR;
}

uint8_t keystore_read_key(uint16_t index, KEY key){
	return keystore_eeprom_read(KEY_MAIN_EEPROM, EEPROM_KEY_OFFSET + (index * sizeof(KEY)), key, sizeof(KEY));
}
uint8_t keystore_read_backup_key(uint16_t index, KEY key){
	return keystore_eeprom_read(KEY_BACKUP_EEPROM, EEPROM_KEY_OFFSET + (index * sizeof(KEY)), key, sizeof(KEY));
}

uint8_t keystore_write_key(uint16_t index, KEY key){
	return keystore_eeprom_write(KEY_MAIN_EEPROM, EEPROM_KEY_OFFSET + (index * sizeof(KEY)), key, sizeof(KEY));
}
uint8_t keystore_write_backup_key(uint16_t index, KEY key){
	return keystore_eeprom_write(KEY_BACKUP_EEPROM, EEPROM_KEY_OFFSET + (index * sizeof(KEY)), key, sizeof(KEY));
}

uint8_t keystore_eeprom_read(uint8_t eeprom, uint16_t offset, uint8_t* buffer, uint8_t length){
	uint8_t ret;
	i2c_master_set_speed(I2C_BITRATE_REG_1M);
	ret = i2c_master_read(eeprom, offset, sizeof(uint16_t), buffer, length);
	if (ret != I2C_SUCCESS) {
		printf_P(PSTR("EEPROM %02x read error %02x, retry...\n"), eeprom, ret);
		ret = i2c_master_read(eeprom, offset, sizeof(uint16_t), buffer, length);
		if (ret != I2C_SUCCESS) printf_P(PSTR("EEPROM %02x read error %02x (fatal)\n"), eeprom, ret);
	}
	return (ret == I2C_SUCCESS);
}

uint8_t keystore_eeprom_write(uint8_t eeprom, uint16_t offset, uint8_t* buffer, uint8_t length){
	uint8_t ret;
	uint8_t write_protect_flag;
	if (eeprom == KEY_MAIN_EEPROM){
		write_protect_flag = EEPROM_WRITE_PROTECT_MAIN;
	} else {
		write_protect_flag = EEPROM_WRITE_PROTECT_BACKUP;
	}
	
	EEPROM_WRITE_PROTECT_PORT &= ~write_protect_flag;
	
	i2c_master_set_speed(I2C_BITRATE_REG_1M);
	ret = i2c_master_write(eeprom, offset, sizeof(uint16_t), buffer, length);
	if (ret != I2C_SUCCESS) {
		printf_P(PSTR("EEPROM %02x Write error %02x, retry...\n"), eeprom, ret);
		ret = i2c_master_write(eeprom, offset, sizeof(uint16_t), buffer, length);
		if (ret != I2C_SUCCESS) printf_P(PSTR("EEPROM %02x write error %02x (fatal)\n"), eeprom, ret);
	}
	
	EEPROM_WRITE_PROTECT_PORT |= write_protect_flag;
	
	return (ret == I2C_SUCCESS);
}