#include "readline.h"

#define USART_NUMBER 1
#define USART_NAME MGMT
#define USART_USE_AS_STDOUT
#include "../../common/serial.h"

uint8_t readline_echo = 1;
char* readline_prompt = "> ";
uint8_t readline_parse_enable = 1;


static char read_buffer[MAX_CMD_LEN+1];
static uint8_t read_len = 0;
static uint8_t discard_newline = 0; //if we just read a \r, ignore \n if it follows

static readline_parsed_cmd_t cmd;

uint8_t readline_parse_command_buffer(void);
void readline_write_prompt(void);
void readline_dump_command(void);
void readline_exec_command(void);

void readline_init(void){
	MGMT_init();
}

void readline_show_welcome_prompt(void){
	printf_P(PSTR("\nAVR command line interface...\n"));
	readline_write_prompt();
}

void readline_write_prompt(void){
	if (readline_prompt) printf(readline_prompt);
}

void readline_set_mode(char* prompt, uint8_t echo, uint8_t parse){
	readline_echo = echo;
	readline_parse_enable = parse;
	readline_prompt = prompt;
}

void readline_process(void){
	if (!MGMT_data_available()) return;
	
	char c = MGMT_read_char();
	
	if (discard_newline && c == '\n') {return;}
	discard_newline = (c == '\r');
	
	if (c == '\r' || c == '\n'){
		MGMT_transmit_char(c); //echo char
		
		if (readline_parse_enable){
			if (readline_parse_command_buffer()){
				readline_exec_command();
			}
		} else {
			read_buffer[read_len] = '\0';
			readline_handle_buffer(read_buffer);
		}
		readline_write_prompt();
		read_len = 0;
	} else if (c == 8){ //backspace
		if (read_len > 0){
			if (readline_echo) {
				MGMT_transmit_char(c); //echo char
				MGMT_transmit_char(' '); //need to emit a space and another del to clean the terminal display
				MGMT_transmit_char(c);  //not sure why del echo does not last, but thats the way it is.
			}
			read_len--;	
		}
	} else if (c >= 0x20) { //don't store control chars in the buffer
		if (read_len < MAX_CMD_LEN){ //don't overflow the read buffer ;)
			if (readline_echo) MGMT_transmit_char(c); //echo char
			
			read_buffer[read_len] = c;
			read_len++;
		}
	}
}

uint8_t readline_find_token_end(char* buffer, uint8_t offset, uint8_t max){
	for (uint8_t i = offset; i <= max; i++){
		if (buffer[i] == ' ' || buffer[i] == '\0' || buffer[i] == '\t'){
			return i;
		}
	}
	return max+1;
}

uint8_t readline_parse_command_buffer(void){
	uint8_t token_index = 0;
	uint8_t offset = 0; //beginning of a token (inclusive)
	uint8_t end_pos = 0; //end of a token (exclusive)
	while(offset < read_len && token_index <= MAX_ARGS){
		end_pos = readline_find_token_end(read_buffer, offset, read_len-1);
		read_buffer[end_pos] = '\0';
		
		
		if (end_pos > offset){
			if (token_index == 0){
				cmd.cmd = read_buffer + offset;
			} else {
				cmd.args[token_index-1] = read_buffer + offset;
			}
			token_index++;
		}
		offset = end_pos + 1;
	}
	
	if (token_index){
		cmd.num_args = token_index - 1;
		return 1;
	} else {
		//Nothing found
		cmd.cmd = read_buffer;
		cmd.cmd[0] = '\0';
		cmd.num_args = 0;
		return 0;
	}
}

void readline_exec_command(void){
	uint8_t i = 0;
	while(readline_commands[i].cmd != 0){
		if (strcmp(readline_commands[i].cmd, cmd.cmd) == 0){
			if (cmd.num_args >= readline_commands[i].args_min && cmd.num_args <= readline_commands[i].args_max){
				readline_commands[i].handler(&cmd);
				return;
			} else {
				printf_P(PSTR("Error: Command '%s' expects %i to %i arguments (supplied: %i)\n"),
				         readline_commands[i].cmd,
				         readline_commands[i].args_min,
				         readline_commands[i].args_max,
				         cmd.num_args);
				return;
			}
		}
		i++;
	}
	printf_P(PSTR("Command not found\n"));
}

void readline_dump_command(void){
	printf_P(PSTR("Command: \"%s\" (len=%i), num_args=%i, args: "), cmd.cmd, strlen(cmd.cmd), cmd.num_args);
	for (uint8_t i = 0; i < cmd.num_args; i++){
		printf_P(PSTR("Arg%i=\"%s\" (len=%i), "),i , cmd.args[i], strlen(cmd.args[i]));
	}
	printf_P(PSTR("\n"));
}