import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;

import javax.smartcardio.Card;
import javax.smartcardio.CardChannel;
import javax.smartcardio.CardException;
import javax.smartcardio.CardTerminal;
import javax.smartcardio.CardTerminals;
import javax.smartcardio.CommandAPDU;
import javax.smartcardio.ResponseAPDU;
import javax.smartcardio.TerminalFactory;

import net.sourceforge.gpj.cardservices.AID;
import net.sourceforge.gpj.cardservices.AIDRegistry;
import net.sourceforge.gpj.cardservices.AIDRegistryEntry;
import net.sourceforge.gpj.cardservices.CapFile;
import net.sourceforge.gpj.cardservices.GPUtil;
import net.sourceforge.gpj.cardservices.GlobalPlatformService;
import net.sourceforge.gpj.cardservices.exceptions.GPMakeSelectableException;


@SuppressWarnings("restriction")
public class JavaCardProgrammer {

	AID securityDomainAID = new AID(new byte[]{(byte)0xA0, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00});
	
    byte[][] keys = {GlobalPlatformService.defaultEncKey,
                     GlobalPlatformService.defaultMacKey,
                     GlobalPlatformService.defaultKekKey};
	
    private int keySet = 0;
    
    private byte[] appletID = new byte[]{0x23, 0x23, 0x00, 0x42, 0x42};
    
    private CardChannel channel;
    private GlobalPlatformService service;
	
	private static final byte CARD_KEY_LENGTH = 64;
	
	public JavaCardProgrammer() throws NoSuchAlgorithmException, CardException{
		try{
			this.channel = getCardChannel();
			this.service = new GlobalPlatformService(securityDomainAID, this.channel);
			this.service.addAPDUListener(this.service);
			
			this.service.open();
			this.service.setKeys(keySet, keys[0], keys[1], keys[2], GlobalPlatformService.DIVER_NONE);
			this.service.openSecureChannel(keySet, 0, GlobalPlatformService.SCP_ANY, GlobalPlatformService.APDU_CLR);
            
	    } catch (CardException e) {
	    	if (e.getCause().getMessage().equalsIgnoreCase("SCARD_E_NO_READERS_AVAILABLE")){
	    		System.out.println("No smart card readers found");
	    	}
	    	throw e;
	    	
	    } catch (NoSuchAlgorithmException e) {
	    	if (e.getCause().getMessage().equalsIgnoreCase("SCARD_E_NO_SERVICE")){
	    		System.out.println("No smart card readers found (PC/SC service not running)");
	    	}
	    	throw e;
	    }
	}
	
	private static CardChannel getCardChannel() throws NoSuchAlgorithmException, CardException{
		TerminalFactory tf = TerminalFactory.getInstance("PC/SC", null);

	    CardTerminals terminals = tf.terminals();

	    System.out.println("Found terminals: " + terminals.list());
	    for (CardTerminal terminal : terminals.list(CardTerminals.State.ALL)) {
	    	Card c = null;
	    	try {
	    		c = terminal.connect("*");
	    	} catch (CardException e) {
	    		if (e.getCause().getMessage().equalsIgnoreCase("SCARD_E_NO_SMARTCARD")) {
	    			System.err.println("No card in reader " + terminal.getName());
	    			continue;
	    		} else
	    			e.printStackTrace();
	    	}
	        System.out.println("Found card in terminal: " + terminal.getName());
	        System.out.println("ATR: " + GPUtil.byteArrayToString(c.getATR().getBytes()));
	        
	        CardChannel channel = c.getBasicChannel();
	        return channel;
	    }
	    return null;
	}
	
	private void uploadCapFile(File capFile) throws GPMakeSelectableException, CardException, IOException{		
        CapFile cap = null;
        
        try{
        	cap = new CapFile(new FileInputStream(capFile));
        } catch (IOException ex){
        	System.out.format("Failed to open CAP file '%s': %s", capFile, ex.toString());
        	throw ex;
        }
        
        int loadSize = GlobalPlatformService.defaultLoadSize;
        service.loadCapFile(cap, true, false, loadSize, false, false);
        
        AID p = cap.getPackageAID();
        for (AID a : cap.getAppletAIDs()) {
            service.installAndMakeSelecatable(p, a, null, (byte)0, null, null);
        }
	}
	
	public void initializeCardApplet(File capFile) throws CardException, IOException{       
        AID packageAID = new AID(new byte[]{(byte)0xCA, (byte)0xFF, (byte)0xEE, (byte)0xC0, (byte)0x1A});


        try{
            service.deleteAID(packageAID, true);
        } catch (CardException ce) {
            System.out.println("Could not delete AID: " + packageAID + ", ignoring.");
            // This is when the applet is not there, ignore
        }
        
       	uploadCapFile(capFile);
	}
	
	public List<AID> readAppletList() throws CardException{
		List<AID> result = new ArrayList<AID>();
		AIDRegistry registry = service.getStatus();
		for (AIDRegistryEntry e : registry){
			if (e.isApplet()){
				result.add(e.getAID());
			}
		}
		return result;
	}
	
	public void dumpAppletList() throws CardException{
		//List applets
        AIDRegistry registry = service.getStatus();
        for (AIDRegistryEntry e : registry) {
            AID aid = e.getAID();
            int numSpaces = (15 - aid.getLength());
            String spaces = "";
            String spaces2 = "";
            for (int i = 0; i < numSpaces; i++) {
                spaces = spaces + "   ";
                spaces2 = spaces2 + " ";
            }
            System.out.print("AID: "
                    + GPUtil.byteArrayToString(aid.getBytes())
                    + spaces
                    + " "
                    + GPUtil.byteArrayToReadableString(aid
                            .getBytes()) + spaces2);
            System.out.format(" %s LC: %d PR: 0x%02X\n", e
                    .getKind().toShortString(), e
                    .getLifeCycleState(), e.getPrivileges());
            for (AID a : e.getExecutableAIDs()) {
                numSpaces = (15 - a.getLength()) * 3;
                spaces = "";
                for (int i = 0; i < numSpaces; i++)
                    spaces = spaces + " ";
                System.out
                        .println("     "
                                + GPUtil.byteArrayToString(a
                                        .getBytes())
                                + spaces
                                + " "
                                + GPUtil
                                        .byteArrayToReadableString(a
                                                .getBytes()));
            }
            System.out.println();
        }
	}
	
	public boolean selectApplet(byte[] aid){
		
		ByteArrayOutputStream bo = new ByteArrayOutputStream();
        try {
            bo.write(aid);
        } catch (IOException e) {
        	e.printStackTrace();
			return false;
        }
        CommandAPDU delete = new CommandAPDU(0x00, 0xA4, 0x04, 0x00, bo.toByteArray());
        ResponseAPDU response;
		try {
			response = service.transmit(delete);
		} catch (IllegalStateException e) {
			e.printStackTrace();
			return false;
		} catch (CardException e) {
			e.printStackTrace();
			return false;
		}
        service.notifyExchangedAPDU(delete, response);
        short sw = (short) response.getSW();
        if (sw !=  GlobalPlatformService.SW_NO_ERROR) {
            System.err.println("Select Applet failed, SW: " + GPUtil.swToString(sw));
            return false;
        }
        return true;
	}
	
	
	public byte[] setCardKey(byte[] key){
		if (! selectApplet(appletID)) return null;
				
		ByteArrayOutputStream bo = new ByteArrayOutputStream();
        try {
            bo.write(key);            //BODY
        } catch (IOException e) {
        	e.printStackTrace();
			return null;
        }
        CommandAPDU delete = new CommandAPDU(0x00, 0x01, 0x00, 0x00, bo.toByteArray(), CARD_KEY_LENGTH);
        ResponseAPDU response;
		try {
			response = service.transmit(delete);
		} catch (IllegalStateException e) {
			e.printStackTrace();
			return null;
		} catch (CardException e) {
			e.printStackTrace();
			return null;
		}
        service.notifyExchangedAPDU(delete, response);
        short sw = (short) response.getSW();
        if (sw !=  GlobalPlatformService.SW_NO_ERROR) {
            System.err.println("Key set failed, SW: " + GPUtil.swToString(sw));
            return null;
        }
        return response.getData();
	}
	
	public byte[] readCardKey(byte[] readKey){
		if (! selectApplet(appletID)) return null;
				
		ByteArrayOutputStream bo = new ByteArrayOutputStream();
        try {
            bo.write(readKey);            //BODY
        } catch (IOException e) {
        	e.printStackTrace();
			return null;
        }
        CommandAPDU delete = new CommandAPDU(0x00, 0x04, 0x00, 0x00, bo.toByteArray(), CARD_KEY_LENGTH);
        ResponseAPDU response;
		try {
			response = service.transmit(delete);
		} catch (IllegalStateException e) {
			e.printStackTrace();
			return null;
		} catch (CardException e) {
			e.printStackTrace();
			return null;
		}
        service.notifyExchangedAPDU(delete, response);
        short sw = (short) response.getSW();
        if (sw !=  GlobalPlatformService.SW_NO_ERROR) {
            System.err.println("Key read failed, SW: " + GPUtil.swToString(sw));
            return null;
        }
        return response.getData();
	}
	
	public boolean clearCardKey(){
		if (! selectApplet(appletID)) return false;
		
        CommandAPDU delete = new CommandAPDU(0x00, 0xFF, 0x00, 0x00);
        ResponseAPDU response;
		try {
			response = service.transmit(delete);
		} catch (IllegalStateException e) {
			e.printStackTrace();
			return false;
		} catch (CardException e) {
			e.printStackTrace();
			return false;
		}
        service.notifyExchangedAPDU(delete, response);
        short sw = (short) response.getSW();
        if (sw !=  GlobalPlatformService.SW_NO_ERROR) {
            System.err.println("Key clear failed, SW: " + GPUtil.swToString(sw));
            return false;
        }
        return true;
	}
	
	public void close() throws CardException{
		channel.close();
	}
	
}
