#include "i2c_master.h"

#include <util/delay.h>

void i2c_master_init(void){
	TWBR = I2C_BITRATE_REG_100K; //set
	TWSR = I2C_PRESCALER;   //bitrate
	TWCR = (1<<TWEN); //enable i2c
}

void i2c_master_set_speed(uint8_t bitrate){
// 	TWCR = 0; //enable i2c
// 	TWBR = bitrate;
// 	TWCR = (1<<TWEN); //enable i2c
}

void i2c_master_reset(void){ //TODO: Is this really the proper way to reset the i2c contoller?
	TWCR = (1 << TWEN) | (1 << TWSTO);
	TWCR = (1<<TWINT); //| (1<<TWEN);
	_delay_ms(1);
	TWCR = 0;
	_delay_ms(1);
	TWCR = (1<<TWEN);
}

/** 
 * Waits for the i2c interface to become ready for the next byte (TWINT is cleared) 
*/
uint8_t i2c_master_wait_ready(void){
	uint32_t c = 0;
	while (!(TWCR & (1<<TWINT))) {
		c++;
		if (c >= I2C_TIMEOUT) return 0;
	}
	return 1;
}


/** send stop signal */
void i2c_master_stop(void){
	i2c_master_wait_ready();
	TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);
	while (TWCR & (1<<TWSTO)) {;} //wait for stop condition to be sent.
}


uint8_t i2c_master_start(uint8_t dest, uint8_t mode){
	TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN); //send start signal
	if (!i2c_master_wait_ready()) return I2C_ERROR_TIMEOUT+1; //wait until start signal has been sent.
	if (((TWSR & I2C_STATUS_MASK) != TW_START) &&
		((TWSR & I2C_STATUS_MASK) != TW_REP_START)) return (I2C_ERROR_START_SEND_FAILD); //check status
	
	TWDR = dest | mode;
	TWCR = (1<<TWINT) | (1<<TWEN);
	if (!i2c_master_wait_ready()) return I2C_ERROR_TIMEOUT+2; //wait until the byte has been sent.
	if ((mode == I2C_MODE_WRITE && ((TWSR & I2C_STATUS_MASK) != TW_MT_SLA_ACK)) ||
		(mode == I2C_MODE_READ  && ((TWSR & I2C_STATUS_MASK) != TW_MR_SLA_ACK))) return (I2C_ERROR_DEST_NOT_ACCEPTED); //check status
		
	return I2C_SUCCESS;
}

/**
 * Sends a single byte on the i2c bus.
 * @param b byte to send
 * @param last if non-zero the byte is expected to be answered by a slave NACK, if zero an ACK is expected.
*/
uint8_t i2c_master_send_byte(uint8_t b){
	TWDR = b;
	TWCR = (1<<TWINT) | (1<<TWEN); //transmit
	if (!i2c_master_wait_ready()) return I2C_ERROR_TIMEOUT; //wait until the byte has been sent.
		
	if ((TWSR & I2C_STATUS_MASK) != TW_MT_DATA_ACK ) return I2C_ERROR_SEND_MESSAGE_TOO_LONG;

	return I2C_SUCCESS;
}

/**
 * Receives a single byte from the i2c bus.
 * @param memory address where to write the byte to.
 * @param last if non-zero a NACK will be returned to the sender to signal the end of transmission.
 **/
uint8_t i2c_master_read_byte(uint8_t* b, uint8_t last){
	//set ACK/NACK Status //TODO: could this be too late?
	TWCR = (1<<TWEN) | (1<<TWINT) | (last ? 0 : (1<<TWEA)); //TODO: Is it a good idea to fiddle with this REG before a read?
	
	if (!i2c_master_wait_ready()) return I2C_ERROR_TIMEOUT; //wait for the byte to arrive
	
	
	*b = TWDR;  //read data
	
	if (last){
		if ((TWSR & I2C_STATUS_MASK) != TW_MR_DATA_NACK) return I2C_ERROR_RECV_MESSAGE_TOO_LONG; //not an ack before the last byte,
	} else {
		if ((TWSR & I2C_STATUS_MASK) != TW_MR_DATA_ACK) return I2C_ERROR_RECV_MESSAGE_TOO_SHORT; //not an ack before the last byte, message too short?
	}
	TWCR = (1<<TWINT) | (1<<TWEN);
	
	return I2C_SUCCESS;
}


/**
 * Writes a buffer to a target
 * @param dest i2c address of the target host
 * @param addr register address at the target host
 * @param length number of bytes to transmit
 * @param buffer pointer to the transmit buffer
 */
uint8_t i2c_master_write(uint8_t dest, uint16_t addr, uint8_t addr_size, uint8_t* buffer, uint8_t length){
	uint8_t ret;
	
	i2c_master_last_error_position = 0x01;
	
	ret = i2c_master_start(dest, I2C_MODE_WRITE); //send dest address and write mode
	if (ret != I2C_SUCCESS) I2C_ERROR(ret); //sends stop and returns the error code
	
	i2c_master_last_error_position = 0x02;
	
	if (addr_size >= 2){
		ret = i2c_master_send_byte(addr >> 8);
		if (ret != I2C_SUCCESS) I2C_ERROR(ret); //sends stop and returns the error code
	}

	ret = i2c_master_send_byte(addr & 0xFF);
	if (ret != I2C_SUCCESS) I2C_ERROR(ret); //sends stop and returns the error code

	i2c_master_last_error_position = 0x03;
	
	for (uint8_t i = 0; i < length; i++){
		ret = i2c_master_send_byte(buffer[i]);
		if (ret != I2C_SUCCESS) I2C_ERROR(ret); //sends stop and returns the error code
		i2c_master_last_error_position = 0xD0+i;
	}
	
	i2c_master_stop();
	
	return I2C_SUCCESS;
}

/**
 * Reads bytes from the target into a buffer
 * @param dest i2c address of the target host
 * @param addr register address at the target host
 * @param length number of bytes to read
 * @param buffer pointer to the receive buffer
 */
uint8_t i2c_master_read(uint8_t dest, uint16_t addr, uint8_t addr_size, uint8_t* buffer, uint8_t length){	
	uint16_t ret;
	
	i2c_master_last_error_position = 0x81;
	
	ret = i2c_master_start(dest, I2C_MODE_WRITE); //send dest address and WRITE mode (address transmit is a write)
	if (ret != I2C_SUCCESS) I2C_ERROR(ret); //error
	
	i2c_master_last_error_position = 0x82;
	
	if (addr_size >= 2){
		ret = i2c_master_send_byte(addr >> 8); //send register address (high)
		if (ret != I2C_SUCCESS) I2C_ERROR(ret); //error
	}

	ret = i2c_master_send_byte(addr & 0xFF);  //send register address (low)
	if (ret != I2C_SUCCESS) I2C_ERROR(ret); //sends stop and returns the error code
	
	i2c_master_last_error_position = 0x83;
	
	if (length == 0){
		i2c_master_stop();
		return I2C_SUCCESS; //nothing more to do.
	}
	/* ----------- READ PHASE ----------- */
	
	i2c_master_last_error_position = 0x84;
	
	ret = i2c_master_start(dest, I2C_MODE_READ); //send dest address and WRITE mode (address transmit is a write)
	if (ret != I2C_SUCCESS) I2C_ERROR(ret); //error
	
	i2c_master_last_error_position = 0x85;
	
	for (uint8_t i = 0; i < length; i++){
		ret = i2c_master_read_byte(&(buffer[i]), i+1 == length);
		if (ret != I2C_SUCCESS) I2C_ERROR(ret); //sends stop and returns the error code
		i2c_master_last_error_position = 0x86+i;
	}
	
	i2c_master_stop();
	
	TWCR = (1<<TWINT);
	
	return I2C_SUCCESS;
}