// #ifndef _SERIAL_H
#include <stdlib.h>
#include <avr/io.h>
#include <util/delay.h>

#ifndef BAUD
	#define BAUD 38400
#endif

#define NONE 0
#define EVEN 1
#define ODD 2
#ifndef PARITY
	#define PARITY NONE
#endif
#ifndef STOPBITS
	#define STOPBITS 1
#endif

#ifndef XPASTE
	//Macro magic to glue #defined names together. See http://groups.google.com/group/comp.lang.c/msg/8c085fe5a8ad1296?hl=en
	#define PASTE(a,b) a ## b
	#define XPASTE(a,b) PASTE(a,b)
#endif

#define ERR_TIMEOUT 0xFFFF

#ifdef USART_NUMBER //MUST defined on devices with multiple USARTs (e.g. ATMega64 or ATMega128)
	
	//Alias the register names depending on which USART we use.
	#define UBRRH          XPASTE(UBRR, XPASTE(USART_NUMBER,H))
	#define UBRRL          XPASTE(UBRR, XPASTE(USART_NUMBER,L))
	#define UCSRA          XPASTE(UCSR, XPASTE(USART_NUMBER,A))
	#define UCSRB          XPASTE(UCSR, XPASTE(USART_NUMBER,B))
	#define UCSRC          XPASTE(UCSR, XPASTE(USART_NUMBER,C))
	#define UCSZ0          XPASTE(UCSZ, XPASTE(USART_NUMBER,0))
	#define UCSZ1          XPASTE(UCSZ, XPASTE(USART_NUMBER,1))
	#define UPM0           XPASTE(UPM,  XPASTE(USART_NUMBER,0))
	#define UPM1           XPASTE(UPM,  XPASTE(USART_NUMBER,1))
	#define USBS           XPASTE(USBS,        USART_NUMBER)
	#define RXEN           XPASTE(RXEN,        USART_NUMBER)
	#define TXEN           XPASTE(TXEN,        USART_NUMBER)
	#define UCSRA          XPASTE(UCSR, XPASTE(USART_NUMBER,A))
	#define UDRE           XPASTE(UDRE,        USART_NUMBER)
	#define UDR            XPASTE(UDR,         USART_NUMBER)
	#define RXC            XPASTE(RXC,         USART_NUMBER)
	#define TXC            XPASTE(TXC,         USART_NUMBER)
	#define U2X            XPASTE(U2X,         USART_NUMBER)
	#define RXCIE          XPASTE(RXCIE,       USART_NUMBER)
	#define USART_RXC_vect XPASTE(USART, XPASTE(USART_NUMBER,_RX_vect))
	
	//If not defined otherwise all function names will be prefixed with "USARTx" where x is the USART number
	#ifndef USART_NAME
		#define USART_NAME XPASTE(USART, USART_NUMBER)
	#endif
#else
	//If not defined otherwise all function names will be prefixed with "USART"
	#ifndef USART_NAME
		#define USART_NAME USART
	#endif
#endif

//Prefix the function names with the USART_NAME
#define FUNC_init                 XPASTE(USART_NAME, _init)
#define FUNC_transmit_char        XPASTE(USART_NAME, _transmit_char)
#define FUNC_flush                XPASTE(USART_NAME, _flush)
#define FUNC_transmit_buffer      XPASTE(USART_NAME, _transmit_buffer)
#define FUNC_transmit_buffer_hex  XPASTE(USART_NAME, _transmit_buffer_hex)
#define FUNC_trash_receive_buffer XPASTE(USART_NAME, _trash_receive_buffer)
#define FUNC_data_available       XPASTE(USART_NAME, _data_available)
#define FUNC_read_char            XPASTE(USART_NAME, _read_char)
#define FUNC_read_char_timeout    XPASTE(USART_NAME, _read_char_timeout)
#define FUNC_read_buffer          XPASTE(USART_NAME, _read_buffer)
#define FUNC_read_buffer_timeout  XPASTE(USART_NAME, _read_buffer_timeout)
#define FUNC_switch_baudrate      XPASTE(USART_NAME, _switch_baudrate)
#define FUNC_switch_mode_transmit XPASTE(USART_NAME, _switch_mode_transmit)
#define FUNC_switch_mode_receive  XPASTE(USART_NAME, _switch_mode_receive)
#define FUNC_enable_rx_interrupt  XPASTE(USART_NAME, _enable_rx_interrupt)
#define FUNC_disable_rx_interrupt XPASTE(USART_NAME, _disable_rx_interrupt)

#ifdef USART_USE_AS_STDOUT //Sets up the serial port as a stream for stdout. With this printf and stuff will write to the serial port.
	#include <stdio.h>
	
	#define FUNC_transmit_stdout XPASTE(USART_NAME,_transmit_stdout)
	
	void FUNC_transmit_char(unsigned char data);
	
	void FUNC_transmit_stdout(char c, FILE* stream){
		FUNC_transmit_char((unsigned char)c);
	}

    static FILE USART_stdout = FDEV_SETUP_STREAM(XPASTE(USART_NAME,_transmit_stdout), NULL, _FDEV_SETUP_WRITE);
	
#endif


#ifdef USART_INTERRUPT_HANDLER	
	#include <avr/interrupt.h>
	
	ISR(USART_RXC_vect){
		USART_INTERRUPT_HANDLER();
	}
	
	void FUNC_enable_rx_interrupt(void){
		UCSRB |= (1 << RXCIE);
	}
	
	void FUNC_disable_rx_interrupt(void){
		UCSRB &= ~(1 << RXCIE);
	}
#endif

void FUNC_init(void){
	//Set bit rate
	#include <util/setbaud.h>
	UBRRH = UBRRH_VALUE;
	UBRRL = UBRRL_VALUE;
	#if USE_2X
		UCSRA |= (1 << U2X);
	#else
		UCSRA &= ~(1 << U2X);
	#endif
	
	//Set frame format
	UCSRC = (1<<UCSZ0) | (1<<UCSZ1) //8 data bits
	#if STOPBITS == 2
		| (1<<USBS)
	#endif
	#if (PARITY == EVEN)
		| (1<<UPM1)
	#endif
	#if (PARITY == ODD)
		| (1<<UPM0) | (1<<UPM1)
	#endif
	#ifdef URSEL
		 | (1<<URSEL)
	#endif
		 ;
	//Enable receiver, transmitter
	UCSRB = (1<<RXEN) | (1<<TXEN);
	
	#ifdef USART_USE_AS_STDOUT
		stdout = &USART_stdout; //assign serial output stream as stdout.
	#endif
}

void FUNC_switch_baudrate(uint32_t baud){
	uint16_t ubrr_value = (((F_CPU) + 8UL * baud) / (16UL * baud) -1UL);
	UBRRH = ubrr_value >> 8;
	UBRRL = ubrr_value & 0xFF;
	while (UCSRA & (1 << RXC)){ //trash receive buffer
		UDR;
	}
}

void FUNC_switch_mode_receive(void){
	while ( !( UCSRA & (1 << TXC)) ) {;} //Wait for transmit complete
	UCSRB &= ~(1<<TXEN);
	UCSRB |=  (1<<RXEN);
	while (UCSRA & (1 << RXC)){UDR;} //trash receive buffer
}
void FUNC_switch_mode_transmit(void){
	UCSRB |=  (1<<TXEN);
	UCSRB &= ~(1<<RXEN);
	while (UCSRA & (1 << RXC)){UDR;} //trash receive buffer
}


void FUNC_transmit_char(unsigned char data){
	//Wait for empty transmit buffer
	while ( !( UCSRA & (1 << UDRE)) ) {;}
	//Put data into buffer, sends the data
	UDR = data;
}

void FUNC_flush(void){
	//Wait for empty transmit buffer
	while ( !( UCSRA & (1 << TXC)) ) {;}
}

void FUNC_transmit_buffer(unsigned char* data, unsigned char length){
	for (unsigned char i = 0; i < length; i++){
		while ( !( UCSRA & (1 << UDRE)) ) {;} //Wait for empty transmit buffer
		UDR = data[i];
	}
}

void FUNC_transmit_buffer_hex(unsigned char* data, unsigned char length){
	unsigned char n;
	unsigned char b;
	for (unsigned char i = 0; i < length; i++){
		b = data[i];
		n = '0' + (b >> 4 & 0xF);
		FUNC_transmit_char((n > '9') ? n + 7 : n);
		n = '0' + (b & 0xF);
		FUNC_transmit_char((n > '9') ? n + 7 : n);
	}
}

void FUNC_trash_receive_buffer(void){
	while ((UCSRA & (1 << RXC)) != 0){
		UDR;
	}
}

inline unsigned char FUNC_data_available(void) {
	return (UCSRA & (1 << RXC)) != 0;
}

unsigned char FUNC_read_char(void) {
	//Wait for data to arrive
	while (!FUNC_data_available()) {;}
	//Return received data
	return UDR;
}

uint16_t FUNC_read_char_timeout(uint16_t timeout) {
	//Wait for data to arrive
	while (!FUNC_data_available() && timeout != 0) {timeout--;_delay_us(100);}
	if (timeout==0) return ERR_TIMEOUT;
	//Return received data
	return UDR;
}



void FUNC_read_buffer(unsigned char* buf, unsigned char length) {
	while (length != 0){
		//Wait for data to arrive
		while (!FUNC_data_available()) {;}
		buf[0] = UDR; //write byte to buffer
		buf++; //go to next byte in buffer
		length--; //one byte less to read
	}
}

uint8_t FUNC_read_buffer_timeout(uint8_t* buf, uint8_t length, uint16_t char_timeout) {
	uint8_t count = 0;
	while (count != length){
		uint16_t ret = FUNC_read_char_timeout(char_timeout);
		if (ret == ERR_TIMEOUT) break;
		buf[count] = ret & 0xFF;
		count++;
	}
	return count;
}