#include "transport.h"

#define USART_NAME USART
#undef USART_USE_AS_STDOUT
#define BAUD 9600
// #define USART_USE_AS_STDOUT //Debugging only!
void USART_recv_interrupt(void);
#define USART_NUMBER TRANSPORT_USART_NUMBER
#define USART_INTERRUPT_HANDLER USART_recv_interrupt
#include "../common/serial.h"

#define SC_MSG_START 0x2B //2B
#define SC_MSG_END   0xD4 //or ~2B
#define SC_INTERRUPT 0xAA
#define SC_ESCAPE 0x5C
#define EC_2B 0x01
#define EC_D4 0x02
#define EC_5C 0x03
#define EC_AA 0x04

#define INTERRUPT_NUM(c) ((c) & 0x0F)
#define INTERRUPT_NONE 0xFF

#define RET_MSG_START (SC_MSG_START << 8)
#define RET_MSG_END   (SC_MSG_END << 8)

interrupt_msg_handler_t interrupt_msg_table[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

#define RECV_IDLE           0x00
#define RECV_START          0x01
#define RECV_BODY           0x02
#define RECV_CHECKSUM       0x03
#define RECV_FINISHED       0x04
#define RECV_ERROR          0x0A
#define RECV_ESCAPE_FLAG    0x40
#define RECV_INTERRUPT_FLAG 0x80
#define RECV_SET_STATE(s) {msg_recv_state = (msg_recv_state & (RECV_ESCAPE_FLAG | RECV_INTERRUPT_FLAG)) | s;}
#define RECV_STATE() (msg_recv_state & ~(RECV_ESCAPE_FLAG | RECV_INTERRUPT_FLAG))
volatile uint8_t  msg_recv_state = RECV_IDLE;
volatile uint8_t  msg_recv_pos = 0;
volatile uint8_t  msg_recv_body_length = 0;
volatile uint16_t msg_recv_checksum = 0;
#define RECV_MAX_LENGTH 255
volatile uint8_t msg_recv_buffer[RECV_MAX_LENGTH];

uint8_t transport_unescape(uint8_t c);

void USART_recv_interrupt(void){
	uint8_t c = USART_read_char();
	if (msg_recv_state & RECV_INTERRUPT_FLAG){
		if (c != INTERRUPT_NONE && interrupt_msg_table[INTERRUPT_NUM(c)]){
			interrupt_msg_table[INTERRUPT_NUM(c)]();
		}
		msg_recv_state &= ~RECV_INTERRUPT_FLAG;
	} else if (c == SC_ESCAPE){
		msg_recv_state |= RECV_ESCAPE_FLAG;
	} else if (c == SC_INTERRUPT){
		msg_recv_state |= RECV_INTERRUPT_FLAG;
	} else if (c == SC_MSG_START){
		RECV_SET_STATE(RECV_START);
		msg_recv_pos = 0;
	} else if (c == SC_MSG_END){
		msg_recv_body_length = msg_recv_pos;
		RECV_SET_STATE(RECV_CHECKSUM);
		msg_recv_pos = 0;
	} else {
		
		if (msg_recv_state & RECV_ESCAPE_FLAG){
			c = transport_unescape(c);
			msg_recv_state &= ~RECV_ESCAPE_FLAG;
		}
		
		if (RECV_STATE() == RECV_START){
			msg_recv_pos = 0;
			RECV_SET_STATE(RECV_BODY);
		}
		
		if (RECV_STATE() == RECV_BODY){
			if (msg_recv_pos < RECV_MAX_LENGTH){
				msg_recv_buffer[msg_recv_pos] = c;
				msg_recv_pos++;
			}
		} else if (RECV_STATE() == RECV_CHECKSUM){
			if (msg_recv_pos == 0) msg_recv_checksum = c;
			if (msg_recv_pos == 1) {
				msg_recv_checksum |= c << 8;
				RECV_SET_STATE(RECV_FINISHED);
			}
			msg_recv_pos++;
		}
	}
}

uint8_t transport_unescape(uint8_t c){
	if (c == EC_2B){
		return 0x2B;
	} else if (c == EC_D4){
		return 0xD4;
	} else if (c == EC_5C){
		return 0x5C;
	} else if (c == EC_AA){
		return 0xAA;
	} else {
		return c;
	}
}

void transport_init(void){
	USART_init();
	USART_enable_rx_interrupt();
	USART_trash_receive_buffer();
}

void transport_set_interrupt_handler(uint8_t interrupt, interrupt_msg_handler_t handler){
	interrupt_msg_table[interrupt] = handler;
}

void transport_flush(void){
	USART_flush();
}

void transport_send_char(uint8_t c){
	if (c == SC_ESCAPE){
		USART_transmit_char(SC_ESCAPE);
		USART_transmit_char(EC_5C);
	} else if (c == SC_MSG_START){
		USART_transmit_char(SC_ESCAPE);
		USART_transmit_char(EC_2B);
	} else if (c == SC_MSG_END){
		USART_transmit_char(SC_ESCAPE);
		USART_transmit_char(EC_D4);
	} else if (c == SC_INTERRUPT){
		USART_transmit_char(SC_ESCAPE);
		USART_transmit_char(EC_AA);
	} else {
		USART_transmit_char(c);
	}
}

/**
 * Sends a message with the given body and length.
 * WARNING: Do not send msgs from an interrupt, this can result in corrupted receive or a reset (interrupt handler must be short).
**/
void transport_send_message(uint8_t* body, uint8_t length){
	uint16_t checksum = 0xFFFF;
	//Send this 3 times. If the remote side is not expecting a message start
	//e.g because it is sleeping in power save mode or because it waits for the
	//end of an other msg that got truncated it will lose at least on of these
	//(or more if the error handling takes time).
	USART_transmit_char(SC_MSG_START);
	USART_transmit_char(SC_MSG_START);
	USART_transmit_char(SC_MSG_START);
	
	for (uint8_t i = 0; i < length; i++){
		transport_send_char(body[i]);
// 		printf(">%02x>", body[i]);
		checksum = _crc16_update(checksum, body[i]);
	}
	USART_transmit_char(SC_MSG_END);
	
	transport_send_char(checksum & 0xFF);
	transport_send_char(checksum >> 8);
	
	USART_flush();
}

void transport_send_interrupt_message(uint8_t interrupt){
	//Dummy interrupt msg to wake the remote side if sleeping
// 	USART_transmit_char(SC_INTERRUPT);
// 	USART_transmit_char(INTERRUPT_NONE);
	
	USART_transmit_char(SC_INTERRUPT);
	USART_transmit_char(INTERRUPT_NUM(interrupt));
}

uint8_t transport_data_available(void){
	return msg_recv_state == RECV_FINISHED;
}

uint16_t transport_receive_message(uint8_t* buffer, uint8_t length, uint16_t timeout){
	
	while (RECV_STATE() != RECV_FINISHED && timeout > 0){
		timeout--;
		_delay_us(100);
	}
	
	if (RECV_STATE() != RECV_FINISHED){
		return TRANSPORT_ERR_TIMEOUT;
	} else if (msg_recv_body_length > length) {
		RECV_SET_STATE(RECV_IDLE);
		return TRANSPORT_ERR_MESSAGE_TOO_LONG;
	} else {
		uint16_t checksum = 0xFFFF;
		for (uint8_t i = 0; i < msg_recv_body_length; i++){
			checksum = _crc16_update(checksum, msg_recv_buffer[i]);
			buffer[i] = msg_recv_buffer[i];
		}
		RECV_SET_STATE(RECV_IDLE);
		
		if (checksum != msg_recv_checksum) return TRANSPORT_ERR_CHECKSUM_FAIL;
		
		return msg_recv_body_length;
	}
}
